{ **************************************************************************** }
{                                                                              }
{ Smart Mobile Studio - Runtime Library                                        }
{                                                                              }
{ Copyright (c) Eric Grange, Creative IT. All rights reserved.                 }
{ Licensed to Optimale Systemer AS.                                            }
{                                                                              }
{ **************************************************************************** }

unit GLS.Vectors;

interface

uses w3c.TypedArray;

type
   Vector2 = array [0..1] of Float;
   Vector3 = array [0..2] of Float;
   Vector4 = array [0..3] of Float;

   Matrix3 = array [0..8] of Float;
   Matrix4 = array [0..15] of Float;

   Quaternion = record
      X, Y, Z, W : Float;
      class function Create(x, y, z, w : Float) : Quaternion;
      function Conjugate : Quaternion;
      function Length : Float;
      function Multiply(const q : Quaternion) : Quaternion; overload;
      function Multiply(const qW : Float) : Quaternion; overload;
      function Normalize: Quaternion;
      function Norm: Float;
      function ToVector3: Vector3;
      function Inverse: Quaternion;
      class function FromEuler(pitch, yaw, roll: Float): Quaternion;
      class function FromAxis(const v: Vector3; angle: Float): Quaternion;
      function Transform(const vec: Vector3): Vector3; overload;
      function Transform(const q: Quaternion): Quaternion; overload;
      function InverseTransform(const vec: Vector3): Vector3; overload;
      function InverseTransform(const q: Quaternion): Quaternion; overload;
      function Add(const vec: Vector3): Quaternion; overload;
      function Add(const Q: Quaternion): Quaternion; overload;
      procedure GetAxisAngle(var axis: Vector3; var angle: Float);
      const Identity : Quaternion = (X: 0; Y: 0; Z: 0; W: 1.0);
      const Zero: Quaternion = (X: 0; Y: 0; Z: 0; W: 0);
   end;

   TVector3Helper = helper for Vector3
      class function Create(x, y, z : Float) : Vector3;

      function Add(const v : Vector3) : Vector3;
      function Sub(const v : Vector3) : Vector3;
      function Multiply(const v: Vector3): Vector3;
      function Dot(const v : Vector3) : Float;
      function Cross(const v : Vector3) : Vector3;
      function Scale(f : Float) : Vector3;
      function Negate : Vector3;
      function Normalize : Vector3;
      function Length : Float;
      function Direction(const dest : Vector3) : Vector3;
      function Lerp(const v : Vector3; f : Float) : Vector3;
      function Dist(const v : Vector3) : Float;
      function Dist2(const v : Vector3) : Float;

      function ToString : String;
      function ToVector4 : Vector4;
      function ToPoint4 : Vector4;

      const NullVector : Vector3 = [0.0, 0.0, 0.0];
      const XVector    : Vector3 = [1.0, 0.0, 0.0];
      const YVector    : Vector3 = [0.0, 1.0, 0.0];
      const ZVector    : Vector3 = [0.0, 0.0, 1.0];
      const XYZVector  : Vector3 = [1.0, 1.0, 1.0];
   end;

   TVector4Helper = helper for Vector4
      class function Create(x, y, z, w : Float) : Vector4;
      class function CreatePoint(x, y, z : Float) : Vector4;
      class function CreateVector(x, y, z : Float) : Vector4;

      function ToVector3 : Vector3;

      const NullVector : Vector4 = [0.0, 0.0, 0.0, 0.0];
      const NullPoint  : Vector4 = [0.0, 0.0, 0.0, 1.0];
      const XVector    : Vector4 = [1.0, 0.0, 0.0, 0.0];
      const YVector    : Vector4 = [0.0, 1.0, 0.0, 0.0];
      const ZVector    : Vector4 = [0.0, 0.0, 1.0, 0.0];
   end;

   TMatrix3Helper = helper for Matrix3
      class function Create(const vx, vy, vz : Vector3) : Matrix3; overload;
      class function CreateRotation(const axis : Vector3; angle : Float) : Matrix3; overload;

      function Multiply(const m : Matrix3) : Matrix3;
      function Scale(const v : Vector3) : Matrix3;
      function Transpose : Matrix3;
      function Inverse : Matrix3;

      function ToFloat32Array : JFloat32Array;
      function ToMatrix4 : Matrix4;

      const Identity : Matrix3 = [1.0, 0.0, 0.0,
                                  0.0, 1.0, 0.0,
                                  0.0, 0.0, 1.0];
   end;

   TMatrix4Helper = helper for Matrix4
      class function Create(const vx, vy, vz : Vector3) : Matrix4; overload;
      class function Create(const vx, vy, vz, vw : Vector3) : Matrix4; overload;
      class function CreateFrustum(left, right, bottom, top, near, far : Float) : Matrix4;
      class function CreatePerspective(fovy, aspect, near, far : Float) : Matrix4;
      class function CreateOrtho(left, right, bottom, top, near, far : Float) : Matrix4;
      class function CreateLookAt(const eye, target, up : Vector3) : Matrix4;

      function Transpose : Matrix4;
      function Multiply(const m : Matrix4) : Matrix4; overload;
      function Multiply(const m : Matrix3) : Matrix4; overload;
      function Translate(const v : Vector3) : Matrix4;
      function Scale(const v : Vector3) : Matrix4; overload;
      function Scale(const v : Float) : Matrix4; overload;
      function RotateX(angle : Float) : Matrix4;
      function RotateY(angle : Float) : Matrix4;
      function RotateZ(angle : Float) : Matrix4;
      function Rotate(const axis : Vector3; angle : Float) : Matrix4;

      function TransformVector(const v : Vector3) : Vector3;
      function TransformPoint(const v : Vector3) : Vector3;
      function Transform(const v : Vector4) : Vector4;

      function ToFloat32Array : JFloat32Array;
      function ToMatrix3 : Matrix3;

      const Identity : Matrix4 = [1.0, 0.0, 0.0, 0.0,
                                  0.0, 1.0, 0.0, 0.0,
                                  0.0, 0.0, 1.0, 0.0,
                                  0.0, 0.0, 0.0, 1.0];
   end;

implementation

// ------------------
// ------------------ Quaternion ------------------
// ------------------

class function Quaternion.Create(x, y, z, w: Float) : Quaternion;
begin
   Result.X := x;
   Result.Y := y;
   Result.Z := z;
   Result.W := w;
end;

function Quaternion.Add(const vec: Vector3): Quaternion;
begin
  Result.X := X + vec[0];
  Result.Y := Y + vec[1];
  Result.Z := Z + vec[2];
  Result.W := W;
end;

function Quaternion.Add(const Q: Quaternion): Quaternion;
begin
  Result.X := X + Q.X;
  Result.Y := Y + Q.Y;
  Result.Z := Z + Q.Z;
  Result.W := W + Q.W;
end;

function Quaternion.Conjugate : Quaternion;
begin
   Result.X := -X;
   Result.Y := -Y;
   Result.Z := -Z;
   Result.W := W;
end;

function Quaternion.InverseTransform(const q: Quaternion): Quaternion;
begin
  Result := Conjugate.Multiply(q.Multiply(Self));
end;

function Quaternion.InverseTransform(const vec: Vector3): Vector3;
var
	vecQuat, resQuat: Quaternion;
begin
	vecQuat.x := vec[0];
	vecQuat.y := vec[1];
	vecQuat.z := vec[2];
	vecQuat.w := 0.0;
  resQuat := InverseTransform(vecQuat);
	Result := Vector3.Create(resQuat.x, resQuat.y, resQuat.z);
end;

function Quaternion.Length : Float;
begin
   Result := Sqrt(X*X + Y*Y + Z*Z + W*W);
end;

function Quaternion.Multiply(const q : Quaternion) : Quaternion;
begin
   Result.X := W * q.X + X * q.W + Y * q.Z - Z * q.Y;
   Result.Y := W * q.Y + Y * q.W + Z * q.X - X * q.Z;
   Result.Z := W * q.Z + Z * q.W + X * q.Y - Y * q.X;
   Result.W := W * q.W - X * q.X - Y * q.Y - Z * q.Z;
end;

function Quaternion.Multiply(const qW : Float) : Quaternion;
begin
   Result.W := W * qW;
   Result.X := X * qW;
   Result.Y := Y * qW;
   Result.Z := Z * qW;
end;

// Convert from Euler Angles
class function Quaternion.FromEuler(pitch, yaw, roll: Float): Quaternion;
var
  sinp, siny, sinr, cosp, cosy, cosr: Float;
begin
	// Basically we create 3 Quaternions, one for pitch, one for yaw, one for roll
	// and multiply those together.
	// the calculation below does the same, just shorter

	sinp := sin(pitch*0.5);
	siny := sin(yaw*0.5);
	sinr := sin(roll*0.5);
	cosp := cos(pitch*0.5);
	cosy := cos(yaw*0.5);
	cosr := cos(roll*0.5);

	Result.x := cosr * cosy * sinp - sinr * siny * cosp;
	Result.y := cosr * siny * cosp + sinr * cosy * sinp;
	Result.z := sinr * cosy * cosp - cosr * siny * sinp;
  Result.w := cosr * cosy * cosp + sinr * siny * sinp;
end;

// Convert from Axis Angle
class function Quaternion.FromAxis(const v: Vector3; angle: Float): Quaternion;
var
  sinAngle: Float;
  vn: Vector3;
begin
	angle *= 0.5;
	vn := v.Normalize;

	sinAngle := sin(angle);

	Result.x := (vn[0] * sinAngle);
	Result.y := (vn[1] * sinAngle);
	Result.z := (vn[2] * sinAngle);
	Result.w := cos(angle);
end;

function Quaternion.Inverse: Quaternion;
begin
  Result := Conjugate.Multiply(1/Norm);
end;

function Quaternion.Norm: Float;
begin
  Result := w * w + x * x + y * y + z * z;
end;

function Quaternion.Normalize: Quaternion;
var
  mag, mag2: Float;
begin
	// Don't normalize if we don't have to
	mag2 := w * w + x * x + y * y + z * z;
	if mag2 <> 0 then
  begin
		mag := sqrt(mag2);
		Result.w := w/mag;
		Result.x := x/mag;
		Result.y := y/mag;
		Result.z := z/mag;
  end;
end;

// Convert to Axis/Angles
procedure Quaternion.GetAxisAngle(var axis: Vector3; var angle: Float);
var
  scale: Float;
begin
	scale := sqrt(x * x + y * y + z * z);
	axis[0] := x / scale;
	axis[1] := y / scale;
	axis[2] := z / scale;
	angle := arccos(w) * 2.0;
end;

// Multiplying a quaternion q with a vector v applies the q-rotation to v
function Quaternion.Transform(const vec: Vector3): Vector3;
var
	vecQuat, resQuat: Quaternion;
begin
	vecQuat.x := vec[0];
	vecQuat.y := vec[1];
	vecQuat.z := vec[2];
	vecQuat.w := 0.0;
  resQuat := Transform(vecQuat);
	Result := Vector3.Create(resQuat.x, resQuat.y, resQuat.z);
end;

function Quaternion.ToVector3: Vector3;
begin
  Result := [X, Y, Z];
end;

function Quaternion.Transform(const q: Quaternion): Quaternion;
begin
  Result := Multiply(q.Multiply(Conjugate));
end;

// ------------------
// ------------------ TVector3Helper ------------------
// ------------------

class function TVector3Helper.Create(x, y, z : Float) : Vector3;
begin
   Result := [ x, y, z ];
end;

function TVector3Helper.Add(const v : Vector3) : Vector3;
begin
   Result := [ Self[0]+v[0], Self[1]+v[1], Self[2]+v[2] ];
end;

function TVector3Helper.Sub(const v : Vector3) : Vector3;
begin
   Result := [ Self[0]-v[0], Self[1]-v[1], Self[2]-v[2] ];
end;

function TVector3Helper.Multiply(const v: Vector3): Vector3;
begin
  Result := [Self[0]*v[0], Self[1]*v[1], Self[2]*v[2]];
end;

function TVector3Helper.Dot(const v : Vector3) : Float;
begin
   Result:=Self[0]*v[0]+Self[1]*v[1]+Self[2]*v[2];
end;

function TVector3Helper.Cross(const v : Vector3) : Vector3;
begin
   Result := [
      Self[1]*v[2]-Self[2]*v[1],
      Self[2]*v[0]-Self[0]*v[2],
      Self[0]*v[1]-Self[1]*v[0]
      ];
end;

function TVector3Helper.Scale(f : Float) : Vector3;
begin
   Result := [ Self[0]*f, Self[1]*f, Self[2]*f ];
end;

function TVector3Helper.Negate : Vector3;
begin
   Result := [ -Self[0], -Self[1], -Self[2] ];
end;

function TVector3Helper.Normalize : Vector3;
var
   n : Float;
begin
   n:=Self[0]*Self[0]+Self[1]*Self[1]+Self[2]*Self[2];
   if n=0 then
      Result:=NullVector
   else begin
      n:=1/Sqrt(n);
      Result:=[Self[0]*n, Self[1]*n, Self[2]*n];
   end;
end;

function TVector3Helper.Length : Float;
begin
   Result:=Sqrt(Self[0]*Self[0]+Self[1]*Self[1]+Self[2]*Self[2]);
end;

function TVector3Helper.Direction(const dest : Vector3) : Vector3;
begin
   Result:=dest.Sub(Self).Normalize;
end;

function TVector3Helper.Lerp(const v : Vector3; f : Float) : Vector3;
begin
   Result := [
      Self[0]+f*(v[0]-Self[0]),
      Self[1]+f*(v[1]-Self[1]),
      Self[2]+f*(v[2]-Self[2])
      ];
end;

function TVector3Helper.Dist(const v : Vector3) : Float;
var
   dx, dy, dz : Float;
begin
   dx:=Self[0]-v[0];
   dy:=Self[1]-v[1];
   dz:=Self[2]-v[2];
   Result:=Sqrt(dx*dx+dy*dy+dz*dz);
end;

function TVector3Helper.Dist2(const v : Vector3) : Float;
var
   dx, dy, dz : Float;
begin
   dx:=Self[0]-v[0];
   dy:=Self[1]-v[1];
   dz:=Self[2]-v[2];
   Result:=dx*dx+dy*dy+dz*dz;
end;

function TVector3Helper.ToString : String;
begin
   Result:=Format('[%f, %f, %f]', [Self[0], Self[1], Self[2]]);
end;

function TVector3Helper.ToVector4 : Vector4;
begin
   Result:=[Self[0], Self[1], Self[2], 0];
end;

function TVector3Helper.ToPoint4 : Vector4;
begin
   Result:=[Self[0], Self[1], Self[2], 1];
end;

// ------------------
// ------------------ TVector4Helper ------------------
// ------------------

class function TVector4Helper.Create(x, y, z, w : Float) : Vector4;
begin
   Result:=[x, y, z, w];
end;

class function TVector4Helper.CreatePoint(x, y, z : Float) : Vector4;
begin
   Result:=[x, y, z, 1];
end;

class function TVector4Helper.CreateVector(x, y, z : Float) : Vector4;
begin
   Result:=[x, y, z, 0];
end;

function TVector4Helper.ToVector3 : Vector3;
begin
   Result:=[Self[0], Self[1], Self[2]];
end;

// ------------------
// ------------------ TMatrix3Helper ------------------
// ------------------

function _DET(a1, a2, b1, b2: Float): Float; overload;
begin
  Result := a1 * b2 - a2 * b1;
end;

function _DET(a1, a2, a3, b1, b2, b3, c1, c2, c3: Float): Float; overload;
begin
  Result :=
    a1 * (b2 * c3 - b3 * c2) -
    b1 * (a2 * c3 - a3 * c2) +
    c1 * (a2 * b3 - a3 * b2);
end;

procedure Adjoint(var M: Matrix3);
var
  a1, a2, a3: Float;
  b1, b2, b3: Float;
  c1, c2, c3: Float;
begin
  a1 := M[0]; a2:= M[1]; a3 := M[2];
  b1 := M[3]; b2:= M[4]; b3 := M[5];
  c1 := M[6]; c2:= M[7]; c3 := M[8];

  M[0]:= _DET(b2, b3, c2, c3);
  M[1]:=-_DET(a2, a3, c2, c3);
  M[2]:= _DET(a2, a3, b2, b3);

  M[3]:=-_DET(b1, b3, c1, c3);
  M[4]:= _DET(a1, a3, c1, c3);
  M[5]:=-_DET(a1, a3, b1, b3);

  M[6]:= _DET(b1, b2, c1, c2);
  M[7]:=-_DET(a1, a2, c1, c2);
  M[8]:= _DET(a1, a2, b1, b2);
end;

function Determinant(const M: Matrix3): Float;
begin
  Result := _DET(M[0], M[3], M[6],
                 M[1], M[4], M[7],
                 M[2], M[5], M[8]);
end;

procedure Scale(var M: Matrix3; Factor: Float);
var
  i: Integer;
begin
  for i := 0 to 8 do
    M[i] := M[i] * Factor;
end;

procedure Invert(var M: Matrix3);
var
  Det: Float;
begin
  Det := Determinant(M);
  if Abs(Det) < 1E-5 then
    M := Matrix3.Identity
  else
  begin
    Adjoint(M);
    Scale(M, 1 / Det);
  end;
end;

class function TMatrix3Helper.Create(const vx, vy, vz : Vector3) : Matrix3;
begin
   Result:=[vx[0], vx[1], vx[2],
            vy[0], vy[1], vy[2],
            vz[0], vz[1], vz[2]];
end;

class function TMatrix3Helper.CreateRotation(const axis : Vector3; angle : Float) : Matrix3;
var
   x, y, z, invLength : Float;
   c, s, omc : Float;
begin
   c:=Cos(angle);
   s:=Sin(angle);
   omc:=1-c; // one minus cosine
   invLength:=1/axis.Length;
   x:=axis[0]*invLength;
   y:=axis[1]*invLength;
   z:=axis[2]*invLength;
   Result:=[ omc*x*x+c,  omc*x*y+z*s,  omc*x*z-y*s,
             omc*x*y-z*s, omc*y*y+c,   omc*y*z+x*s,
             omc*x*z+y*s, omc*y*z-x*s, omc*z*z+c ];
end;

function TMatrix3Helper.Multiply(const m : Matrix3) : Matrix3;
begin
   Result := [
      Self[0]*m[0] + Self[3]*m[1] + Self[6]*m[2],
      Self[1]*m[0] + Self[4]*m[1] + Self[7]*m[2],
      Self[2]*m[0] + Self[5]*m[1] + Self[8]*m[2],
      Self[0]*m[3] + Self[3]*m[4] + Self[6]*m[5],
      Self[1]*m[3] + Self[4]*m[4] + Self[7]*m[5],
      Self[2]*m[3] + Self[5]*m[4] + Self[8]*m[5],
      Self[0]*m[6] + Self[3]*m[7] + Self[6]*m[8],
      Self[1]*m[6] + Self[4]*m[7] + Self[7]*m[8],
      Self[2]*m[6] + Self[5]*m[7] + Self[8]*m[8]
      ];
end;

function TMatrix3Helper.Scale(const v : Vector3) : Matrix3;
begin
   Result := [
      Self[0]*v[0], Self[1]*v[0], Self[2]*v[0],
      Self[3]*v[1], Self[4]*v[1], Self[5]*v[1],
      Self[6]*v[2], Self[7]*v[2], Self[8]*v[2]
      ];
end;

function TMatrix3Helper.Transpose : Matrix3;
begin
   Result := [
      Self[0], Self[3], Self[6],
      Self[1], Self[4], Self[7],
      Self[2], Self[5], Self[8]
      ];
end;

function TMatrix3Helper.Inverse : Matrix3;
(*
var
   b0, b1, b2 : Float;
   det, invDet : Float;
*)
begin
  Result := Self;
  Invert(Result);
(*
   b0:=Self[8]*Self[4]-Self[5]*Self[7];
   b1:=Self[8]*Self[6]-Self[3]*Self[7];
   b2:=Self[7]*Self[3]-Self[4]*Self[6];

   det:=Self[0]*b0+Self[1]*b1+Self[2]*b2;
   invDet:=1/det;

   Result := [
      b0*invDet,
      (Self[2]*Self[7]-Self[8]*Self[1])*invDet,
      (Self[5]*Self[1]-Self[2]*Self[4])*invDet,
      b1*invDet,
      (Self[8]*Self[0]-Self[2]*Self[6])*invDet,
      (Self[2]*Self[3]-Self[5]*Self[0])*invDet,
      b2*invDet,
      (Self[1]*Self[6]-Self[7]*Self[0])*invDet,
      (Self[4]*Self[0]-Self[1]*Self[3])*invDet
      ];
*)
end;

function TMatrix3Helper.ToFloat32Array : JFloat32Array;
begin
   asm
      @Result = new Float32Array(@Self);
   end;
end;

function TMatrix3Helper.ToMatrix4 : Matrix4;
begin
   Result := [
      Self[0], Self[1], Self[2], 0,
      Self[3], Self[4], Self[5], 0,
      Self[6], Self[7], Self[8], 0,
            0,       0,       0, 1
      ];
end;

// ------------------
// ------------------ TMatrix4Helper ------------------
// ------------------

class function TMatrix4Helper.Create(const vx, vy, vz : Vector3) : Matrix4;
begin
   Result:=[vx[0], vx[1], vx[2], 0,
            vy[0], vy[1], vy[2], 0,
            vz[0], vz[1], vz[2], 0,
                0,     0,     0, 1];
end;

class function TMatrix4Helper.Create(const vx, vy, vz, vw : Vector3) : Matrix4;
begin
   Result:=[vx[0], vx[1], vx[2], 0,
            vy[0], vy[1], vy[2], 0,
            vz[0], vz[1], vz[2], 0,
            vw[0], vw[1], vw[2], 1];
end;

class function TMatrix4Helper.CreateFrustum(left, right, bottom, top, near, far : Float) : Matrix4;
begin
   Result:=[ near*2/(right-left), 0, 0, 0,
             0, near*2/(top-bottom), 0, 0,
             (right+left)/(right-left), (top+bottom)/(top-bottom), (far+near)/(near-far), -1,
             0, 0, (2*far*near)/(near-far), 0];
end;

class function TMatrix4Helper.CreatePerspective(fovy, aspect, near, far : Float) : Matrix4;
var
   top, right : Float;
begin
   top:=near*Tan(DegToRad(fovy*0.5));
   right:=top*aspect;
   Result:=CreateFrustum(-right, right, -top, top, near, far);
end;

class function TMatrix4Helper.CreateOrtho(left, right, bottom, top, near, far : Float) : Matrix4;
begin
   Result:=[ 2/(top-left), 0, 0, 0,
             0, 2/(top-bottom), 0, 0,
             0, 0, -2/(far-near), 0,
             -(left+right)/(right-left), -(top+bottom)/(top-bottom), -(far+near)/(far-near), 1];
end;

class function TMatrix4Helper.CreateLookAt(const eye, target, up : Vector3) : Matrix4;
var
   dir, right, fixup, offset : Vector3;
begin
   dir:=eye.Direction(target);
   right:=dir.Cross(up).Normalize;
   fixup:=right.Cross(dir).Normalize;
   offset:=[-right.Dot(eye), -fixup.Dot(eye), -dir.Dot(eye)];
   Result:=Create(right, fixup, dir, offset);
end;

function TMatrix4Helper.Transpose : Matrix4;
begin
   Result:=[Self[00], Self[04], Self[08], Self[12],
            Self[01], Self[05], Self[09], Self[13],
            Self[02], Self[06], Self[10], Self[14],
            Self[03], Self[07], Self[11], Self[15]];
end;

function TMatrix4Helper.Multiply(const m : Matrix4) : Matrix4;
begin
   Result := [
      m[00]*Self[00] + m[01]*Self[04] + m[02]*Self[08] + m[03]*Self[12],
      m[00]*Self[01] + m[01]*Self[05] + m[02]*Self[09] + m[03]*Self[13],
      m[00]*Self[02] + m[01]*Self[06] + m[02]*Self[10] + m[03]*Self[14],
      m[00]*Self[03] + m[01]*Self[07] + m[02]*Self[11] + m[03]*Self[15],
      m[04]*Self[00] + m[05]*Self[04] + m[06]*Self[08] + m[07]*Self[12],
      m[04]*Self[01] + m[05]*Self[05] + m[06]*Self[09] + m[07]*Self[13],
      m[04]*Self[02] + m[05]*Self[06] + m[06]*Self[10] + m[07]*Self[14],
      m[04]*Self[03] + m[05]*Self[07] + m[06]*Self[11] + m[07]*Self[15],
      m[08]*Self[00] + m[09]*Self[04] + m[10]*Self[08] + m[11]*Self[12],
      m[08]*Self[01] + m[09]*Self[05] + m[10]*Self[09] + m[11]*Self[13],
      m[08]*Self[02] + m[09]*Self[06] + m[10]*Self[10] + m[11]*Self[14],
      m[08]*Self[03] + m[09]*Self[07] + m[10]*Self[11] + m[11]*Self[15],
      m[12]*Self[00] + m[13]*Self[04] + m[14]*Self[08] + m[15]*Self[12],
      m[12]*Self[01] + m[13]*Self[05] + m[14]*Self[09] + m[15]*Self[13],
      m[12]*Self[02] + m[13]*Self[06] + m[14]*Self[10] + m[15]*Self[14],
      m[12]*Self[03] + m[13]*Self[07] + m[14]*Self[11] + m[15]*Self[15]
   ];
end;

function TMatrix4Helper.Multiply(const m : Matrix3) : Matrix4;
begin
   Result := [
      m[00]*Self[00] + m[01]*Self[04] + m[02]*Self[08],
      m[00]*Self[01] + m[01]*Self[05] + m[02]*Self[09],
      m[00]*Self[02] + m[01]*Self[06] + m[02]*Self[10],
      m[00]*Self[03] + m[01]*Self[07] + m[02]*Self[11],
      m[03]*Self[00] + m[04]*Self[04] + m[05]*Self[08],
      m[03]*Self[01] + m[04]*Self[05] + m[05]*Self[09],
      m[03]*Self[02] + m[04]*Self[06] + m[05]*Self[10],
      m[03]*Self[03] + m[04]*Self[07] + m[05]*Self[11],
      m[06]*Self[00] + m[07]*Self[04] + m[08]*Self[08],
      m[06]*Self[01] + m[07]*Self[05] + m[08]*Self[09],
      m[06]*Self[02] + m[07]*Self[06] + m[08]*Self[10],
      m[06]*Self[03] + m[07]*Self[07] + m[08]*Self[11],
      Self[12],
      Self[13],
      Self[14],
      Self[15]
   ];
end;

function TMatrix4Helper.Translate(const v : Vector3) : Matrix4;
begin
   Result:=[Self[00], Self[01], Self[02], Self[03],
            Self[04], Self[05], Self[06], Self[07],
            Self[08], Self[09], Self[10], Self[11],
            Self[00]*v[0] + Self[04]*v[1] + Self[08]*v[2] + Self[12],
            Self[01]*v[0] + Self[05]*v[1] + Self[09]*v[2] + Self[13],
            Self[02]*v[0] + Self[06]*v[1] + Self[10]*v[2] + Self[14],
            Self[03]*v[0] + Self[07]*v[1] + Self[11]*v[2] + Self[15]];
end;

function TMatrix4Helper.Scale(const v : Vector3) : Matrix4;
begin
   Result:=[Self[00]*v[0], Self[01]*v[0], Self[02]*v[0], Self[03]*v[0],
            Self[04]*v[1], Self[05]*v[1], Self[06]*v[1], Self[07]*v[1],
            Self[08]*v[2], Self[09]*v[2], Self[10]*v[2], Self[11]*v[2],
            Self[12],      Self[13],      Self[14],      Self[15]];
end;

function TMatrix4Helper.Scale(const v : Float) : Matrix4;
begin
  Result := Scale([v,v,v]);
end;

function TMatrix4Helper.RotateX(angle : Float) : Matrix4;
var
   c, s : Float;
begin
   c:=Cos(angle);
   s:=Sin(angle);
   Result:=[Self[00], Self[01], Self[02], Self[03],
            Self[04]*c+Self[08]*s, Self[05]*c+Self[09]*s, Self[06]*c+Self[10]*s, Self[07]*c+Self[11]*s,
            Self[08]*c-Self[04]*s, Self[09]*c-Self[05]*s, Self[10]*c-Self[06]*s, Self[11]*c-Self[07]*s,
            Self[12], Self[13], Self[14], Self[15]];
end;

function TMatrix4Helper.RotateY(angle : Float) : Matrix4;
var
   c, s : Float;
begin
   c:=Cos(angle);
   s:=Sin(angle);
   Result:=[Self[00]*c-Self[08]*s, Self[01]*c-Self[09]*s, Self[02]*c-Self[10]*s, Self[03]*c-Self[11]*s,
            Self[04], Self[05], Self[06], Self[07],
            Self[08]*c+Self[00]*s, Self[09]*c+Self[01]*s, Self[10]*c+Self[02]*s, Self[11]*c+Self[03]*s,
            Self[12], Self[13], Self[14], Self[15]];
end;

function TMatrix4Helper.RotateZ(angle : Float) : Matrix4;
var
   c, s : Float;
begin
   c:=Cos(angle);
   s:=Sin(angle);
   Result:=[Self[00]*c+Self[04]*s, Self[01]*c+Self[05]*s, Self[02]*c+Self[06]*s, Self[03]*c+Self[07]*s,
            Self[04]*c-Self[00]*s, Self[05]*c-Self[01]*s, Self[06]*c-Self[02]*s, Self[07]*c-Self[03]*s,
            Self[08], Self[09], Self[10], Self[11],
            Self[12], Self[13], Self[14], Self[15]];
end;

function TMatrix4Helper.Rotate(const axis : Vector3; angle : Float) : Matrix4;
begin
   Result:=Self.Multiply(Matrix3.CreateRotation(axis, angle));
end;

function TMatrix4Helper.TransformVector(const v : Vector3) : Vector3;
begin
   Result := [
      Self[0]*v[0] + Self[4]*v[1] + Self[ 8]*v[2],
      Self[1]*v[0] + Self[5]*v[1] + Self[ 9]*v[2],
      Self[2]*v[0] + Self[6]*v[1] + Self[10]*v[2]
      ];
end;

function TMatrix4Helper.TransformPoint(const v : Vector3) : Vector3;
begin
   Result := [
      Self[0]*v[0] + Self[4]*v[1] + Self[ 8]*v[2] + Self[12],
      Self[1]*v[0] + Self[5]*v[1] + Self[ 9]*v[2] + Self[13],
      Self[2]*v[0] + Self[6]*v[1] + Self[10]*v[2] + Self[14]
      ];
end;

function TMatrix4Helper.Transform(const v : Vector4) : Vector4;
begin
   Result := [
      Self[0]*v[0] + Self[4]*v[1] + Self[ 8]*v[2] + Self[12]*v[3],
      Self[1]*v[0] + Self[5]*v[1] + Self[ 9]*v[2] + Self[13]*v[3],
      Self[2]*v[0] + Self[6]*v[1] + Self[10]*v[2] + Self[14]*v[3],
      Self[3]*v[0] + Self[7]*v[1] + Self[11]*v[2] + Self[15]*v[3]
      ];
end;

function TMatrix4Helper.ToFloat32Array : JFloat32Array;
begin
   asm
      @Result=new Float32Array(@Self);
   end;
end;

function TMatrix4Helper.ToMatrix3 : Matrix3;
begin
   Result := [
      Self[0], Self[1], Self[2],
      Self[4], Self[5], Self[6],
      Self[8], Self[9], Self[10]
      ];
end;

